@@ -2,8 +2,8 @@
 
 Copyright (C) 1986, 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994,
     1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
-    2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014 Massachusetts
-    Institute of Technology
+    2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016,
+    2017, 2018, 2019 Massachusetts Institute of Technology
 
 This file is part of MIT/GNU Scheme.
 
@@ -102,13 +102,14 @@ USA.
 	  (error "Illegal LAPOPT pattern - must end with opcode"
 		 (reverse pattern)))
       (let ((key (caar pattern)))
-	(hash-table/put! *rules* key
-			 (cons rule (hash-table/get *rules* key '()))))))
+	(hash-table-set! *rules* key
+			 (cons rule
+			       (hash-table-ref/default *rules* key '()))))))
   name)
 
 (define (find-rules instruction)
-  (hash-table/get *rules* (car instruction) '()))
-  
+  (hash-table-ref/default *rules* (car instruction) '()))
+
 ;; Rules are tried in the reverse order in which they are defined.
 ;;
 ;; Rules are matched against the LAP from the bottom up.
